# FaultNVC
Earthquake Fault Plane Extractor through Point-Cloud Normal Vector Clustering of Hypocenter Distributions

[![License](https://img.shields.io/badge/License-BSD_3--Clause-blue.svg)](https://opensource.org/licenses/BSD-3-Clause)

Copyright (c) 2024, National Institute of Advanced Industrial Science and Technology (AIST). All rights reserved.

## Introduction

The package `FaultNVC` extracts planar fault alignments in hypocenter distributions through hierarchical clustering of hypocenter positions and point cloud normal vectors. `FaultNVC` is a Python package built upon a robust class-based architecture. This package is under development, so any bug reports and suggestions are welcome!

The following instruction shows how to setup and run `FaultNVC` for extracting intricate fault planes from hypocenter distributions. [The Open-File Report of Geological Survey of Japan](https://www.gsj.jp/publications/pub/openfile/openfile0759.html) or a document in `manual` includes detailed explanations for this tool.

## Installation

You first need to unzip the downloaded file. The following instructions assume that the undipped package is placed in the `$HOME/FaultNVC` directory. 

We recommend using a `conda` environment to establish the virtual environment. You need to install `FaultNVC` using `pip` locally. In the instructions below, `nvc` is used as an example name for the environment. 

-  [Recommended] `conda/mamba`
    You may replace `mamba` with `conda` in the following instruction
    ```
    $ cd $HOME/FaultNVC      # move to the unzipped directory
    # mamba create -n nvc python=3.11  # python 3.10+ required 
    $ mamba activate nvc     # do not forget to activate the environment 
    (nvc) $ mamba env update -f ./environment_dep.yml  # installs required packages
    (nvc) $ mamba install pytest       # pytest for installation test (skippable)
    ```
    Then install `FaultNVC`
    ```
    (nvc) $ pip install .        # local installation
    (nvc) $ mamba list FaultNVC  # check its installation
    ```
    This is optional, but you may install optional packages
    ```
    (nvc) $ mamba env update -f ./environment_opt.yml  # installs optional packages
    ```

- [Alternative] `pip` on `venv`
    1. Check if an existing python has version 3.10+
        ```
        $ python -V
        ```
        If not, install and specify Python v3.10+, such as `python3.10`.
    1. Create an environment using `venv`
        ```
        $ mkdir $HOME/venv_src   # Any name is OK
        $ cd $HOME/venv_src
        $ python -m venv nvc     # Specify python3.10 or python3.11 if needed
        $ source nvc/bin/activate
        ```
    1. Install locally
        ```
        (nvc) $ cd $HOME/FaultNVC   # move to the unzipped directory
        (nvc) $ pip install .    # install FaultNVC and other packages
        (nvc) $ pip show FaultNVC  # check the installation

        # pytest for installation test (skippable)
        (nvc) $ pip install pytest
        ```

### Test your installation

Run test files

```
(nvc) $ (cd tests && pytest)
```


### Requirements
`FaultNVC` requires Python 3.10+ and following packages:
- [HDBSCAN](https://github.com/scikit-learn-contrib/hdbscan) (`hdbscan`)
    - Different from [`sklearn.cluster.HDBSCAN`](https://scikit-learn.org/stable/modules/generated/sklearn.cluster.HDBSCAN.html)
- NumPy (`numpy`)
- Numba (`numba`)
- Matplotlib (`matplotlib`)
- scikit-learn (`sklearn`)
- pandas (`pandas`)
- pyproj (`pyproj`)
- joblib (`joblib`)

Following packages are optional:
- [Cartopy](https://github.com/SciTools/cartopy) (`cartopy`)
- Plotly (`plotly`)
- [Colorcet](https://github.com/holoviz/colorcet) (`colorcet`)
- tqdm (`tqdm`)


## How to use

### Preface
The class `FaultHDBSCAN` has numerous methods, performing all the necessary processes for this analysis. The input data is three-dimensional point-cloud data (north, east, and depth), such as hypocenter distributions. The general steps are:
1. Initialize `FaultHDBSCAN`
1. Load point-cloud data and compute point-cloud normal vectors
1. Execute HDBSCAN (clustering) under various parameter sets

### Compute point-cloud normal vectors

```python
import numpy as np
import pandas as pd

from fnvc import FaultHDBSCAN

# Sample dataset (random values)
df_input = pd.read_csv(
    '$HOME/FaultNVC/examples/sample1/points.txt', 
    sep=r'\s+', header=None, names=('E','N','D'),
)

# Parameters for computing point-cloud normal vectors
max_neighbors_normal_vector = 150 
min_neighbors_normal_vector = 20
max_dist_neighbors = 2.0
min_planarity = 1.25 

# Initialize FaultHDBSCAN
clusterer = FaultHDBSCAN(
    max_neighbors_normal_vector=max_neighbors_normal_vector,
    min_neighbors_normal_vector=min_neighbors_normal_vector,
    max_dist_neighbors=max_dist_neighbors,
    min_planarity=min_planarity,
)

# Load point-cloud data and compute point-cloud normal vectors
clusterer.calc_normal_vector(
    df_input,
    id_points=df_input.index.values,
    kw_points=dict(N='N', E='E', D='D'), # convert colnames to (N,E,D)
    standardize=True, # standardize positions
    max_workers=2, # parallel computing
)
```

The method `.plot_events()` plots normal vectors or apparent strike

```python
# Normal vector
fig = clusterer.plot_events(
    includes_normal=True,
    color_normal=True,
    mode='2d', # "2d", "3d", and "cartopy"
    cbar_rect=[0.15,0.15,0.2,0.2]
)

# Apparent strike
fig = clusterer.plot_events(
    includes_normal=False,
    color_normal=False,
    cbar_rect=[0.15,0.15,0.2,0.2]
)
```
![Normal vectors](./figures/normal.png)

### Execute HDBSCAN

The method `.fit_predict()` executes clustering and extracts fault planes under the specified parameter set: 
- `min_samples` ($m_\mathrm{pts}$) $-$ Number of nearest neighbors to calculate a core distance
- `min_cluster_size` ($m_\mathrm{clSize}$) $-$ Minimum size to be considered a cluster
- `cluster_selection_epsilon` ($\varepsilon$) $-$ Maximum distance to merge two clusters in the condensed dendrogram

Candidate values for $m_\mathrm{pts}$ and $m_\mathrm{clSize}$ can be selected through trial and error to ensure that the resulting clusters reflect primary plane structures or include minor clusters, setting `cluster_selection_epsilon` to `0.0`. It is advisable to start with a large value for $m_\mathrm{pts}$ and $m_\mathrm{clSize}$ and then search for a smaller value that allows the desired outcome. However, choosing `min_samples` and `min_cluster_size` might not be straightforward. You may simplify the clustering by setting $m_\mathrm{clSize}=m_\mathrm{pts}$ (Campello et al., 2013; McInnes et al., 2017), depending on the case. 

```python
# key 0
clusterer.fit_predict(
    min_cluster_size=5,
    min_samples=5,
    cluster_selection_epsilon=0.000,
)

# key 1
clusterer.fit_predict(
    min_cluster_size=5,
    min_samples=5,
    cluster_selection_epsilon=0.300,
)

# key 2
clusterer.fit_predict(
    min_cluster_size=3,
    min_samples=2,
    cluster_selection_epsilon=0.000,
)
```

The method `.show_computed_params()` returns keys of the parameter sets:
```
    min_cluster_size  min_samples  cluster_selection_epsilon
0                  5            5                      0.000
1                  5            5                      0.300
2                  3            2                      0.000
```
If you specify `key=0`, the methods return the result of the specified parameter set.
The method `.plot_planes()` creates a figure of extracted fault planes:
```python
# max_angle_diff: maximum median angle offset of point-cloud normal vectors and fault normal
# min_points: minimum number of points to plot a plane
fig = clusterer.plot_faults(
    key=1,
    mode='2d', # "2d", "3d", and "plotly"
    max_angle_diff=30.0,   # default to 90.0
    min_points=20,         # default to 5
)
```
![Normal vectors](./figures/planes.png)

### Output results

If you want to output dataframes or save the results:
```python
# Return the result as dataframes:
# results_clustering: dataframe for extracted fault parameters
# output_events: dataframe for hypocenter cluster numbers 
results_clustering, output_events = clusterer.output_planes(key=1)

# Or save as CSV files
filename = "path/to/samplefile0"
clusterer.save_model(filename=filename, key=1)
# path/to/samplefile0_planes.csv -> results_clustering
# path/to/samplefile0_events.csv -> output_events
```

### A script for `sample1`

You can test the procedure above by running a python script
```terminal
(nvc) $ python ./examples/sample1/sample1.py --min_samples 5 --mode 2d --max_angle_diff 30.0 --min_points 10
```
Optional arguments can be found by `$python sample1.py -h`.

## Terms of use

- Please cite this Open-File Report and our preprint (Sawaki et al., 2025) if you use the package
- Any bug reports and suggestions are welcome!

### Use cases

- Sawaki, Y., Sato, Y., and Uchide, T., 2025. FaultNVC: Earthquake Fault Plane Extractor through Point-Cloud Normal Vector Clustering for Hypocenter Distributions. Open-File Report of Geological Survey of Japan, AIST, no. 759, https://www.gsj.jp/publications/pub/openfile/openfile0759.html
- Sawaki, Y., Shiina, T., Sagae, K., Sato, Y., Horikawa, H., Miyakawa, A., Imanishi, K., & Uchide, T., 2025. Fault Geometries of the 2024 Mw 7.5 Noto Peninsula Earthquake from Hypocenter-Based Hierarchical Clustering of Point-Cloud Normal Vectors [Preprint], *Earth and Space Science Open Archive*.

## Contributions

### Core team

- **Y. Sawaki**[![ORCID](https://orcid.org/sites/default/files/images/orcid_16x16.png)](https://orcid.org/0000-0002-4043-3391) - Principle developer
- Y. Sato[![ORCID](https://orcid.org/sites/default/files/images/orcid_16x16.png)](https://orcid.org/0000-0002-2794-9596) - Developed the two-step clustering method
- T. Uchide[![ORCID](https://orcid.org/sites/default/files/images/orcid_16x16.png)](https://orcid.org/0000-0001-5106-1186) - Project administrator

### Additional contributors

- K. Sagae, A.Mpuang, T. Shiina, and H. Horikawa - Code review

## Notes

- This package has been tested on Ubuntu 22.04. 
- The authors do not assume any responsibility for any issues or consequences that may arise from the use of this package.

## Acknowledgments

- Comments from Ryota Hino and Satoshi Ide improved the method.
- This study was supported by MEXT Project for Seismology toward Research Innovation with Data of Earthquake (STAR-E) Grant Number JPJ010217.

## License

This project is licensed under the BSD 3-Clause License, see [`LICENSE`](./LICENSE) for details.
