# Sample 1


```python
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from fnvc import FaultHDBSCAN

try:
    import colorcet as cc
    cmap = 'cet_CET_D7_r'
except ImportError:
    cmap = 'viridis_r'
```

## Dataset


```python
df_input = pd.read_csv('./points.txt', sep=r'\s+', header=None, names=('E','N','D'))
df_fault = pd.read_json('./fault_params.json', orient='index')
```


```python
fig, ax = plt.subplots(figsize=[9,8])

df_input.sort_values('D', ascending=False).plot.scatter(
    x='E',y='N', s=12, ec='none',
    c='D', cmap=cmap, norm=plt.Normalize(0,15), 
    ax=ax, alpha=0.7
)

for ser in df_fault.itertuples():
    box_arr = np.array(ser.box_arr)
    ax.plot(*box_arr[[1,2,3,0]][:,:2].T, c='k',ls='--')
    ax.plot(*box_arr[[0,1]][:,:2].T, c='k',lw=3)

ax.set_aspect('equal')
ax.set(xlabel='Easting [km]',ylabel='Northing [km]',)
cax = fig.get_children()[2]
cax.set(ylabel='Depth [km]')
cax.invert_yaxis()
fig.suptitle('Synthetic hypocenters', y=0.95);
```


    
![png](./Sample1_files/output_4_0.png)
    


## Initialization


```python
max_neighbors_normal_vector = 150
min_neighbors_normal_vector = 20
max_dist_neighbors = 2.0
min_samples = 5
min_cluster_size = 5
min_planarity = 1.25
```


```python
clusterer = FaultHDBSCAN(
    max_neighbors_normal_vector=max_neighbors_normal_vector,
    min_neighbors_normal_vector=min_neighbors_normal_vector,
    max_dist_neighbors=max_dist_neighbors,
    min_planarity=min_planarity,
)

clusterer.calc_normal_vector(
    df_input,
    id_points=df_input.index.values,
    max_workers=2,
)
```




    array([[ 0.53889511,  0.84224939, -0.01442299],
           [-0.53023484, -0.84501342, -0.06930612],
           [-0.48165783, -0.87196429, -0.08765843],
           ...,
           [-0.68778652,  0.66225994, -0.29725658],
           [-0.4805139 , -0.87464935, -0.06399152],
           [ 0.88258095,  0.29547636, -0.36571108]])



## Normal vectors


```python
fig = clusterer.plot_events(
    mode='2d',  # '3d', 'cartopy'
    includes_normal=True,
    color_normal=True,
    cbar_rect=[0.15,0.15,0.2,0.2], 
    # view_kw=dict(azim=250, elev=30)
)
fig.axes[0].text(
    0.02,0.02,
    r"$k_{\rm NV} = $"+f"{max_neighbors_normal_vector:3d}\n"+
    r"$r_{\rm NV} = $"+f"{max_dist_neighbors:4.1f} [km]",
    transform=fig.axes[0].transAxes
);
```


    
![png](./Sample1_files/output_9_0.png)
    



```python
clusterer.fit_predict(
    min_cluster_size=min_cluster_size,
    min_samples=min_samples,
    cluster_selection_epsilon=0.0,
) # 0
```




    0




```python
fig = clusterer.plot_faults(
    0, 
    mode='2d',  # '3d', 'plotly'
    size=2,
    max_angle_diff=90,
    min_points=5,
)

fig.axes[0].text(
    0.02,0.02,
    r"$m_{\rm pts} = $"+f"{2:2d}\n"+
    r"$m_{\rm clSize} = $"+f"{3:2d}\n"+
    fr"$\varepsilon$ = {0.0:.3f}"+"\n"+
    r"$\theta^{\rm EV}$ = "+f"{90:4.1f}$\degree$\n"+
    r"$m^{\rm EV}$ = "+f"{5:3d}",
    transform=fig.axes[0].transAxes
);
```


    
![png](./Sample1_files/output_11_0.png)
    


## Compare `eps`


```python
max_angle_diff = 30 #90
min_points = 20 #5
```


```python
fig, axs = plt.subplots(nrows=2, ncols=2, figsize=[10,11])
axs = axs.flatten()

for k, eps in enumerate([0.0,0.2,0.3,0.8]):
    if k > 0:
        k = clusterer.fit_predict(
            min_cluster_size=min_cluster_size,
            min_samples=min_samples,
            cluster_selection_epsilon=eps,
        )
    clusterer.plot_faults(
        k, mode='2d', 
        size=2,
        max_angle_diff=max_angle_diff,
        min_points=min_points,
        ax=axs[k], fig=fig
    )
    axs[k].text(
        0.02,0.02,
        r"$m_{\rm pts} = $"+f"{min_samples:2d}\n"+
        r"$m_{\rm clSize} = $"+f"{min_cluster_size:2d}\n"+
        fr"$\varepsilon$ = {eps:.3f}"+"\n"+
        r"$\theta^{\rm EV}$ = "+f"{max_angle_diff:4.1f}$\degree$\n"+
        r"$m^{\rm EV}$ = "+f"{min_points:3d}",
        transform=axs[k].transAxes
    )
```


    
![png](./Sample1_files/output_14_0.png)
    



```python
# small m_pts and m_clSize
fig = clusterer.fit_predict(
    min_cluster_size=3,
    min_samples=2,
    cluster_selection_epsilon=0.0,
) # 4

fig = clusterer.plot_faults(
    4, 
    mode='2d',  # '3d', 'plotly'
    size=2,
    max_angle_diff=90,
    min_points=5,
)

fig.axes[0].text(
    0.02,0.02,
    r"$m_{\rm pts} = $"+f"{2:2d}\n"+
    r"$m_{\rm clSize} = $"+f"{3:2d}\n"+
    fr"$\varepsilon$ = {0.0:.3f}"+"\n"+
    r"$\theta^{\rm EV}$ = "+f"{90:4.1f}$\degree$\n"+
    r"$m^{\rm EV}$ = "+f"{5:3d}",
    transform=fig.axes[0].transAxes
);
```


    
![png](./Sample1_files/output_15_0.png)
    



```python
clusterer.show_computed_params()
```

       min_cluster_size  min_samples  cluster_selection_epsilon
    0                 5            5                        0.0
    1                 5            5                        0.2
    2                 5            5                        0.3
    3                 5            5                        0.8
    4                 3            2                        0.0



```python
# # if plotly installed
# fig = clusterer.plot_faults(
#     key=2, mode='plotly', 
#     max_angle_diff=15.0,
#     min_points=20,
# )

# # synthetic fault model
# for ser in df_fault.itertuples():
#     box_arr = np.array(ser.box_arr)
#     fig.add_mesh3d(
#         **{k: v for k, v in zip('xyz',box_arr.T)},
#         hovertemplate=(
#             '%{text}<br>'
#             f'L: {ser.L} [km]<br>'
#             f'W: {ser.W} [km]<br>'
#             f'strike: N{ser.strike}°E<br>'
#             f'dip: {ser.dip}°<br>'
#             'x: %{x:7.3f}<br>'
#             'y: %{y:7.3f}<br>'
#             'z: %{z:7.3f}'
#         ),
#         delaunayaxis='y',
#         text=[f'C{ii}' for ii in range(1,5)],
#         opacity=0.5, 
#         name=f"F{ser.Index}",
#         showlegend=True,
#         visible='legendonly',
#         color='gray',
#     )
    
# fig.show()

## save as html
# fig.write_html('planes.html')
```

## Output model

### Return the result as dataframes:

- `results_clustering`: dataframe for extracted fault parameters
- `output_events`: dataframe for hypocenter cluster numbers


```python
results_clustering, output_events = clusterer.output_planes(key=2)
```

### Save as CSV fiiles


```python
# clusterer.save_model("./output", key=2)
```
