# sample1.py
# See LICENSE file for full license details.
# Copyright (c) 2024, National Institute of Advanced Industrial Science and Technology (AIST). All rights reserved.


from pathlib import Path
from argparse import ArgumentParser

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from fnvc import FaultHDBSCAN
try:
    import colorcet as cc
    cmap = 'cet_CET_D7_r'
except ImportError:
    cmap = 'viridis_r'


if __name__ == "__main__":
    parser = ArgumentParser(description='Test code for sample 1')
    parser.add_argument('--knv', type=int, default=150, help='The number of nearest neighbors')
    parser.add_argument('--rnv', type=float, default=2.0, help='The maximum distance to nearest neighbors')
    parser.add_argument('--min_samples', type=int, default=5, help='Number of nearest neighbors to calculate a core distance')
    parser.add_argument('--min_cluster_size', type=int, default=None, help='Minimum size to be considered a cluster')
    parser.add_argument('--epsilon', nargs='+', type=float, default=[0.0,0.2,0.3,0.8], help='a list of integers')
    parser.add_argument('--mode', type=str, default='2d', help='mode for plot')
    parser.add_argument('--max_angle_diff', type=float, default=90.0, help='Threshold for angle difference')
    parser.add_argument('--min_points', type=int, default=5, help='Minimum points to plot as a fault plane')
    args = parser.parse_args()

    mode = args.mode
    max_angle_diff = args.max_angle_diff
    min_points = args.min_points

    df_input = pd.read_csv(
        Path(__file__).parent/'points.txt', 
        sep=r'\s+', header=None, names=('E','N','D')
    )
    df_fault = pd.read_json(
        Path(__file__).parent/'fault_params.json', 
        orient='index',
    )

    fig1, ax = plt.subplots(figsize=[9,8])
    df_input.sort_values('D', ascending=False).plot.scatter(
        x='E',y='N', s=12, ec='none',
        c='D', cmap=cmap, norm=plt.Normalize(0,15), 
        ax=ax, alpha=0.7
    )
    for ser in df_fault.itertuples():
        box_arr = np.array(ser.box_arr)
        ax.plot(*box_arr[[1,2,3,0]][:,:2].T, c='k',ls='--')
        ax.plot(*box_arr[[0,1]][:,:2].T, c='k',lw=3)
    ax.set_aspect('equal')
    ax.set(xlabel='Easting [km]',ylabel='Northing [km]',)
    cax = fig1.get_children()[2]
    cax.set(ylabel='Depth [km]')
    cax.invert_yaxis()
    fig1.suptitle('Synthetic hypocenters', y=0.95)


    max_neighbors_normal_vector = args.knv  # 150
    min_neighbors_normal_vector = 20
    max_dist_neighbors = args.rnv  # 2.0
    min_samples = args.min_samples
    min_cluster_size = max(
        min_samples if args.min_cluster_size is None else args.min_cluster_size,
        3
    )
    min_planarity = 1.25

    clusterer = FaultHDBSCAN(
        max_neighbors_normal_vector=max_neighbors_normal_vector,
        min_neighbors_normal_vector=min_neighbors_normal_vector,
        max_dist_neighbors=max_dist_neighbors,
        min_planarity=min_planarity,
    )

    clusterer.calc_normal_vector(
        df_input,
        id_points=df_input.index.values,
        max_workers=2,
    )

    fig2 = clusterer.plot_events(
        mode=mode if mode != 'plotly' else '3d',  # '3d', 'cartopy'
        includes_normal=True,
        color_normal=True,
        cbar_rect=[0.15,0.15,0.2,0.2], 
        fig_kw=dict(figsize=(9,8), dpi=100)
    )
    if mode == '2d':
        fig2.axes[0].text(
            0.02,0.02,
            r"$k_{\rm NV} = $"+f"{max_neighbors_normal_vector:3d}\n"+
            r"$r_{\rm NV} = $"+f"{max_dist_neighbors:4.1f} [km]",
            transform=fig2.axes[0].transAxes
        )


    k = clusterer.fit_predict(
        min_cluster_size=min_cluster_size,
        min_samples=min_samples,
        cluster_selection_epsilon=0.0,
    )


    fig3, axs = plt.subplots(nrows=2, ncols=2, figsize=[10,11])
    axs = axs.flatten()

    for k, eps in enumerate(args.epsilon):
        if k > 0:
            k = clusterer.fit_predict(
                min_cluster_size=min_cluster_size,
                min_samples=min_samples,
                cluster_selection_epsilon=eps,
            )
        clusterer.plot_faults(
            k, mode='2d', 
            max_angle_diff=max_angle_diff,
            min_points=min_points,
            ax=axs[k], fig=fig3
        );
        axs[k].text(
            0.02,0.02,
            r"$m_{\rm pts} = $"+f"{min_samples:2d}\n"+
            r"$m_{\rm clSize} = $"+f"{min_cluster_size:2d}\n"+
            fr"$\varepsilon$ = {eps:.3f}"+"\n"+
            r"$\theta^{\rm EV}$ = "+f"{max_angle_diff:4.1f}$\degree$\n"+
            r"$m^{\rm EV}$ = "+f"{min_points:3d}",
            transform=axs[k].transAxes
        )

    clusterer.show_computed_params()

    fig4 = clusterer.plot_faults(
        key=2, 
        mode=(mode if mode != 'cartopy' else '3d'), 
        max_angle_diff=max_angle_diff,
        min_points=min_points,
    )
    if mode == "plotly":
        fig4.show()
    else:
        ax = fig4.axes[0]
        if mode == '2d':
            ax.text(
                0.02,0.02,
                r"$m_{\rm pts} = $"+f"{min_samples:2d}\n"+
                r"$m_{\rm clSize} = $"+f"{min_cluster_size:2d}\n"+
                fr"$\varepsilon$ = {args.epsilon[2]:.3f}"+"\n"+
                r"$\theta^{\rm EV}$ = "+f"{max_angle_diff:4.1f}$\degree$\n"+
                r"$m^{\rm EV}$ = "+f"{min_points:3d}",
                transform=ax.transAxes
            )

    plt.show()

    # output model (k=2)
    print('Creating `output_events.csv` and `output_faults.csv` ...')
    clusterer.save_model(filename=Path(__file__).parent/"output", key=2)

