# fnvc/clustering.py
# See LICENSE file for full license details.
# Copyright (c) 2024, National Institute of Advanced Industrial Science and Technology (AIST). All rights reserved.


import sys
import os
from pathlib import Path
import string
from datetime import datetime, timedelta
from dataclasses import dataclass, field, asdict
from functools import partial
from itertools import product
import warnings

import numpy as np
from numba import njit, prange
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
import pandas as pd
from joblib import Parallel, delayed
from sklearn.neighbors import NearestNeighbors
from hdbscan import HDBSCAN

from .utils import (
    normalize_min_max, 
    compute_each_normal_vector, 
    compute_fault_param,
    fix_aspectratio,
    _initialize_CSS4_COLORS
)

try:
    import colorcet as cc
    DEFAULT_CYCLIC_CMAP = 'cet_CET_C6_r'
except ImportError:
    DEFAULT_CYCLIC_CMAP = 'twilight'

COLNAMES_PLANE = (
    'index_cluster', 'num_clusters',
    'strike', 'dip',
    'length', 'width',
    *[f'{c}_{X}' for c,X in product(['xi','eta','n'],'NED')],
    'variance_normal', 'variance_ratio_normal', 'planarity',
    'median_angle_diff',
    *[f'R{i}_{X}' for i,X in product(range(4),'NED')], # fault rectangle
)

COLNAMES_EVENT = (
    'index_cluster', 
    'N','E','D','nn_org','ne_org','nd_org',
    'pc3_ratio', 'planarity',
    'catalog_index', 'id'
)

@dataclass
class BaseClustering():
    """Class for clustering events by normal vectors.
    The normal vector is computed from N nearest neighbors.
    
    Note that we use the coordinate system of
    [North, East, Down] positive. Thus, the input data of
    [X,Y,Z] is transformed into [Y,X,Z] (+Z & +D for downward). 

    
    Parameters
    ----------
    max_neighbors_normal_vector: int
        Maximum number of neighbors for calculating normal
    min_neighbors_normal_vector: int
        Minimum number of neighbors for calculating normal
    max_dist_neighbors: float = None
        Maximum distance of neighbors for calculating normal
    min_planarity: float = 1.25
        Minimum planarity (λ2/λ3) of a point to be used
    
    """
    
    max_neighbors_normal_vector: int = 100
    min_neighbors_normal_vector: int = 5
    max_dist_neighbors: float = None
    min_planarity: float = 1.25

    def __post_init__(self):
        self._cluster_colors = _initialize_CSS4_COLORS()


    def calc_normal_vector(
            self, 
            all_points: np.ndarray|pd.DataFrame, 
            id_points=None,
            kw_points=dict(
                N='N',
                E='E',
                D='D',
                id='id',
                origin_time='origin_time'
            ),
            beg_time=None,
            end_time=None,
            standardize=True,
            max_workers=None,
        ):
        """Calculate the normal vector for the plane formed by
        the top two principle-component vectors.

        Each dataset composes of `n_neighbors` points retrieved
        using the nearest neighbor method with KDtree algorithm.

        Parameters:
        ----------
        all_points: ndarray
            Array of all data points.
            1D array of [X, Y, Z] for each row.
        id_points: list or ndarray
            Event IDs (digits)
        kw_points: dict
            Component names.
        beg_time: None
            Start time of used events 
        end_time: None
            End time of used events 
        standardize: default to True
            If standardizes event positions for clustering
        max_workers: int
            Maximum number of workers in `joblib.Parallel` (`n_jobs`)

        Returns
        -------
        self.normal_vectors_kneighbors: ndarray
            Point cloud normal vectors
        """

        def _check_timedelta(value):
            if isinstance(value, dict):
                return timedelta(**value)
            elif isinstance(value, timedelta):
                return value
            elif value is None:
                return None
            else:
                raise TypeError('beg_time or end_time must be a `timedelta` or `dict`.')

        if (beg_time is None) ^ (end_time is None):
            raise ValueError('Both beg_time and end_time must be specified.')
        
        flg_time = beg_time is not None
        
        beg_time = _check_timedelta(beg_time)
        end_time = _check_timedelta(end_time)

        if isinstance(all_points, np.ndarray):
            if all_points.shape[1] < 3:
                raise Exception("`all_points` must have three columns of XYZ.")
            else:
                # Convert the input XYZ to NED (D=Z)
                all_points = all_points[:,[1,0,2]]

        elif isinstance(all_points, pd.DataFrame):
            kw_points=dict(
                N='N',
                E='E',
                D='D',
                id='id',
                origin_time='origin_time'
            ) | kw_points

            df_points = all_points.copy()
            kw_id = kw_points['id']
            if kw_id in df_points.columns:
                id_points = df_points.get(kw_id)
            all_points = df_points[[kw_points.get(key) for key in 'NED']].values

        if id_points is None:
            id_points = pd.Series(np.random.randint(0, 10**9, size=len(all_points)), name='id')
            while id_points.duplicated().any():
                dup_ids = id_points[id_points.duplicated()].index
                id_points.loc[dup_ids] = np.random.randint(0, 10**9, size=len(dup_ids))
            self.id_points = id_points
        else:
            if isinstance(id_points, (pd.DataFrame,pd.Series)):
                self.id_points = id_points
                if isinstance(id_points, pd.Series) and (self.id_points.name != 'ID'):
                    self.id_points = self.id_points.rename('id')
            else:
                self.id_points = pd.Series(id_points, name='id')
        self.id_points_org = self.id_points.copy()

        if flg_time:

            def _find_knn_time(
                j, 
                df=df_points, 
                N=self.max_neighbors_normal_vector, 
                beg_time=beg_time, 
                end_time=end_time, 
                kw_points=kw_points,
            ):
                tb = df.origin_time[j] + beg_time
                te = df.origin_time[j] + end_time

                NED = [kw_points.get(key) for key in 'NED']

                arr_r = (((
                    df.loc[j,NED]-df.query(f'@tb<={kw_points.get("origin_time")}<=@te&index!=@j')[NED]
                )**2).sum(axis=1)**0.5).sort_values()[:N]

                indices = arr_r.index.values
                arr_r = arr_r.astype(float).values

                return j, arr_r, indices
            
            if max_workers is None:
                max_workers = -1
            
            _results = Parallel(n_jobs=max_workers)([delayed(_find_knn_time)(j) for j in df_points.index])

            _arr_dist = dict()
            _index_kneighbors = dict()

            for j, r, i in _results:
                _arr_dist[j] = r
                _index_kneighbors[j] = i
                
            self.arr_dist = [_arr_dist.get(k) for k in range(len(_arr_dist))]
            self.index_kneighbors = [_index_kneighbors.get(k) for k in range(len(_index_kneighbors))]

            del _results, _arr_dist, _index_kneighbors

        else:
            knn_model = NearestNeighbors(
                algorithm='kd_tree', 
                n_neighbors=self.max_neighbors_normal_vector,
            ).fit(all_points) 

            # Distance & indices of the K-neighbors of points
            # index_kneighbors[i,j]: j-th index of the K-neighbors of i-th point
            self.arr_dist, self.index_kneighbors = knn_model.kneighbors(all_points)
        
        if self.max_dist_neighbors:
            bad_indices_less_points = [j for j, dist in enumerate(self.arr_dist) if np.sum(dist<=self.max_dist_neighbors)<self.min_neighbors_normal_vector]

            arr_dist = self.arr_dist.copy()
            index_kneighbors = self.index_kneighbors.copy()
            self.arr_dist = [] 
            self.arr_dist_bad = []
            self.index_kneighbors = []
            self.index_kneighbors_bad = []

            for j, (dist, indices) in enumerate(zip(arr_dist, index_kneighbors)):
                if j in bad_indices_less_points:
                    self.arr_dist_bad.append(dist)
                    self.index_kneighbors_bad.append(indices)
                else:
                    cond = dist <= self.max_dist_neighbors
                    self.arr_dist.append(dist[cond])
                    self.index_kneighbors.append(indices[cond])
            
            self.normal_vectors_kneighbors = np.array([
                self.compute_each_normal_vector(all_points[ith_kneighbors], returns_EVR=True)
                for ith_dist, ith_kneighbors in zip(self.arr_dist, self.index_kneighbors)
            ])
            
            self.bad_indices_less_points = bad_indices_less_points
            self.excluded_points = all_points[bad_indices_less_points]
            all_points = np.delete(all_points, bad_indices_less_points, 0)
            self.id_points.drop(index=bad_indices_less_points, inplace=True)
            
        else:
            self.normal_vectors_kneighbors = np.array([
                self.compute_each_normal_vector(all_points[ith_kneighbors], returns_EVR=True)
                for ith_kneighbors in self.index_kneighbors
            ])
        
        self.id_points = self.id_points.reset_index()

        self.explained_variance_ratio = self.normal_vectors_kneighbors[:,3:]
        self.planarity = self.explained_variance_ratio[:,1] / self.explained_variance_ratio[:,2]
        self.boolarr_good_nv = self.planarity >= self.min_planarity
        self.normal_vectors_kneighbors = self.normal_vectors_kneighbors[:,:3]
        
        self.NED_normal_vectors_kneighbors = np.hstack([
            all_points, 
            self.normal_vectors_kneighbors
        ])

        ## Standardization of 3-D locations
        if standardize:
            std_locations = np.sqrt(np.sum(np.var(all_points,axis=0)))
            self.feature_vector = np.hstack([
                (all_points-all_points.mean(axis=0))/std_locations, 
                self.normal_vectors_kneighbors
            ])
        else:
            self.feature_vector = np.hstack([
                all_points, 
                self.normal_vectors_kneighbors
            ])        
        
        self.XYZ_normal_vectors_kneighbors = self.NED_normal_vectors_kneighbors[:,[1,0,2,4,3,5]]

        return self.normal_vectors_kneighbors

    @staticmethod
    def compute_each_normal_vector(locations, returns_EVR=False):
        """Compute the normal vector (PC3 eigen vector)

        Parameters:
        -----------
        locations: ndarray
            Array of event locations (NEZ)
        """
        
        if len(locations) < 3:
            return np.zeros(3) * np.nan

        return compute_each_normal_vector(locations, returns_EVR=returns_EVR)
        
    def output_planes(self, labels, min_points=5):
        """Output clustering results of extracted planes and events
        """

        results_clustering = []
        output_events = []
        output_cluster_colors = []

        good_points = self.NED_normal_vectors_kneighbors[self.boolarr_good_nv]
        good_ids = self.id_points[self.boolarr_good_nv]

        label_excluded = []
        for i_label in sorted(set(labels)):
            if i_label < 0:
                label_excluded.append(i_label)
                continue

            points = good_points[labels==i_label]
            if len(points) < min_points:
                label_excluded.append(i_label)
                continue

            (
                box_arr, 
                fault_length, 
                fault_width, 
                normal_vector,
                strike_vector, 
                dip_vector, 
                variance, 
                planarity,
                pc3_variance_ratio
            ) = compute_fault_param(points[:,:3])

            strike = np.round(np.rad2deg(np.angle(strike_vector[:2]@np.array([1,1j])))%360, 2)
            dip = np.round(np.rad2deg(np.arccos(-normal_vector[2])), 2)

            δφ = np.abs(points[:,3:]@normal_vector)
            δφ = np.rad2deg(np.arccos(np.minimum(δφ, 1.0)))  # affected by rounding
            median_angle_diff = np.median(δφ)

            results_clustering.append([
                i_label, 
                len(points),
                strike, dip,  # in degrees
                fault_length, fault_width,
                *np.round(strike_vector,5),
                *np.round(dip_vector,5),
                *np.round(normal_vector,5),
                np.round(variance[2],5),  # variance in normal
                np.round(pc3_variance_ratio,5),
                np.round(planarity,5),
                round(median_angle_diff,2),
                *np.round(box_arr.flatten(),3)
            ])

            ## event information
            # for point_k, id_k in zip(points, good_ids[labels==i_label].values):
            for point_k, ids_k in zip(points, good_ids[labels==i_label].itertuples()):
                ind_nv = ids_k.Index
                output_events.append([
                    i_label,
                    *np.round(point_k,4),  # [N,E,D,nN,nE,nD]
                    self.explained_variance_ratio[ind_nv,2],
                    self.planarity[ind_nv],
                    ids_k.index, ids_k.id,
                    # *id_k,  # 'catalog_index', 'id'
                ])

        for i_label in label_excluded:
            points = good_points[labels==i_label]

            ## event information
            for point_k, ids_k in zip(points, good_ids[labels==i_label].itertuples()):
                ind_nv = ids_k.Index
                output_events.append([
                    i_label,
                    *np.round(point_k,4),
                    self.explained_variance_ratio[ind_nv,2],
                    self.planarity[ind_nv],
                    ids_k.index, ids_k.id,
                ])

        # [-2] points excluded by planarity control for KNN-PCA
        for ind_nv in np.where(self.boolarr_good_nv==False)[0]:
            point_k = self.XYZ_normal_vectors_kneighbors[ind_nv]
            output_events.append([
                -2,
                *np.round(point_k,4),
                self.explained_variance_ratio[ind_nv,2],
                self.planarity[ind_nv],
                *self.id_points.loc[ind_nv].values
            ])

        # [-9] points not calculated normal vectors
        if "excluded_points" in dir(self):
            for point_k, id_k in zip(self.excluded_points, self.bad_indices_less_points):
                output_events.append([
                    -9,
                    *np.round(point_k,4),
                    *[np.nan]*5,
                    id_k, self.id_points_org.loc[id_k]
                ])

        results_clustering = pd.DataFrame(
            results_clustering,
            columns=COLNAMES_PLANE,
        ).set_index('index_cluster')
        output_events = pd.DataFrame(
            output_events,
            columns=COLNAMES_EVENT,
        ).set_index('index_cluster').sort_values(by='id')
        output_events[['pc3_ratio','planarity']] = output_events[['pc3_ratio','planarity']].round(4)

        return results_clustering, output_events
    
    
    def plot_faults(
            self,
            data=None,
            mode="2d",
            max_angle_diff=30., 
            max_PC3_ratio=1/3, 
            min_planarity=1.0,
            min_dipping=0.,
            min_L=0.0,
            min_W=0.0,
            size=2., 
            kw_planes_mpl=dict(c='k', lw=1., ls='--', alpha=0.5),
            cmap=None, 
            cone_scale=1.0, 
            plotly_layout=dict(),
            fig_kw=None,
            subplot_kw=None,
            fig=None, 
            ax=None,
            results_clustering=None,
            output_events=None,
            **view_kwargs
        ):
        """Plot extracted fault planes

        Parameters
        ----------
        data: ndarray, default to None
            Data points to be ploted
            If None, original input data will be plotted.
        mode: str, default to `'2d'`
            Type of projection. Select one from
            `'2d'`, `'3d'`, and `'plotly'`
        max_angle_diff: float, default to 30.0
            Maximum median angle difference between
            fault normal and point-cloud normal vectors.
            Specify 90.0 to plot all the clusters 
        max_PC3_ratio: float, default to 1/3
            Maximum PC3 ratio to plot
        min_points: int
            Mimimum number of points in a cluster
        min_dipping: float
            Mimimum dipping angle
        min_L: float
            Mimimum length of a fault plane
        min_W: float
            Mimimum width of a fault plane
        size: float
            Size of scatter plot
        kw_planes_mpl: dict
            kw dict in mpl plot for planes
        cmap: str or mcolors.Colormap, default to None
            cmap of colorbar. 
            If None, `DEFAULT_CYCLIC_CMAP` is used.
        cone_scale: float
            Size of normal vectors
        plotly_layout: dict
            `layout` argument for `plotly.graph_objects.Figure()`
        fig_kw: dict
             kwargs for `plt.figure()`
        subplot_kw: dict
            kwargs for `fig.add_subplot()`
        fig: plt.Figure, default to None
            Specify an existing Figure object or newly create
        ax: plt.Axes, default to None
            Specify an existing AXes object or 
            newly create from the figure object.

        Returns
        -------
        fig: plt.Figure
        """
        if mode not in ["2d","3d","plotly"]:
            raise ValueError(f"Invalid input '{mode}' for `mode`.")

        if data is None:
            data = self.XYZ_normal_vectors_kneighbors

        if not isinstance(cmap, mcolors.Colormap):
            if isinstance(cmap, str):
                cmap = plt.get_cmap(cmap)
            else:
                cmap = plt.get_cmap('jet')

        max_nvv = np.cos(np.deg2rad(min_dipping))


        ###########################################
        ### Prepare a figure for each plotting mode

        if mode == 'plotly':
            try:
                import plotly.graph_objects as go
            except ImportError:
                raise ImportError(
                    "`plotly` is not properly installed. Try\n"
                    "`conda install plotly` or `pip install plotly`"
                )
            
            plotly_layout = dict(
                width=1200, height=800, 
                title='Extracted fault planes', 
                showlegend=True,
                legend=dict(x=1, y=0.5, orientation='v')
            )| plotly_layout

            fig = go.Figure(
                data=(
                    go.Scatter3d(
                        **{k: v for k, v in zip('xyz', data[:,:3].T)},
                        mode='markers',
                        marker=dict(size=0.01, opacity=0.0, color='rgba(0,0,0,0)'),
                        showlegend=False,
                    ), # dummy plot
                    go.Scatter3d(
                        **{k: v for k, v in zip('xyz', data[:,:3].T)},
                        mode='markers',
                        marker=dict(size=size, opacity=0.8, color='gray'),
                        name='Seismicity',
                    ),
                ),
                layout=plotly_layout,
            )

            fig.update_scenes(
                xaxis_title='Easting [km]',
                yaxis_title='Northing [km]',
                zaxis_title='Depth [km]',
                xaxis=dict(nticks=10,),
                yaxis=dict(nticks=10,),
                zaxis=dict(nticks=10,autorange="reversed"),
                camera=dict(
                    up=dict(x=0, y=0, z=-1),
                    # center=dict(x=0, y=0, z=0),
                    eye=dict(x=0, y=-2., z=1.5)
                ),
            )

        else:  # "2d" or "3d"
            if isinstance(ax, plt.Axes):
                fig = ax.figure
            if fig is None:
                default_fig_kw = dict(figsize=(9,8), dpi=100)
                if fig_kw:
                    default_fig_kw.update(fig_kw)
                fig = plt.figure(**default_fig_kw)

            if not isinstance(ax, plt.Axes):
                default_subplot_kw = dict()
                if subplot_kw:
                    default_subplot_kw.update(subplot_kw)
                if mode == "3d":
                    default_subplot_kw['projection'] = "3d"
                ax = fig.add_subplot(**default_subplot_kw)

            if mode == '2d':
                ax.scatter(*data[:,:2].T,s=size,fc='0.5',ec='none',alpha=0.5)

                ax.set(
                    xlabel='Easting [km]', ylabel='Northing [km]',
                    aspect='equal'
                )
            
            elif mode == '3d':
                ax.scatter(*data[:,:3].T,s=size,fc='0.5',ec='none',alpha=0.5)

                ax.set(
                    xlabel='Easting [km]', ylabel='Northing [km]', zlabel='Depth [km]',
                )
                ax.invert_zaxis()
                ax.set_aspect('equal')
                ax.set_title('Extracted fault planes',y=0.90)

                ax.view_init(**view_kwargs) # elev=20, azim=360-60, roll=0

        ### Preparation of a figure
        ###########################################
            
        ### Plot fault planes
        self.__plot_common(
            mode=mode,
            results_clustering=results_clustering,
            output_events=output_events,
            max_PC3_ratio=max_PC3_ratio,
            min_planarity=min_planarity,
            max_nvv=max_nvv,
            max_angle_diff=max_angle_diff,
            min_L=min_L, min_W=min_W,
            kw_planes_mpl=kw_planes_mpl,
            cone_scale=cone_scale,
            cmap=cmap,
            fig=fig,
            ax=[None, ax][mode != 'plotly']
        )

        # fix aspect ratio for plotly plot
        if mode == "plotly":
            fix_aspectratio(fig, only_visible=True)

        return fig
    
    def __plot_common(
            self, 
            mode, 
            results_clustering,
            output_events,
            max_PC3_ratio, 
            min_planarity,
            max_nvv, 
            max_angle_diff, 
            min_L, 
            min_W, 
            kw_planes_mpl,
            cone_scale, 
            cmap, 
            fig,
            ax=None,
        ):

        def col_plotly(strike, cmap=cmap):
            # Cyclic color for 0π: E-W; π/2: N-S
            return mcolors.rgb2hex(cmap(strike/360)[:3]).upper()
        
        n_clusters = len(set(results_clustering.index))
        
        _normalize_min_max = partial(
            normalize_min_max, 
            min=results_clustering.R0_D.min(),
            max=results_clustering.R0_D.max(),
        )

        # only for mpl plot
        if (alpha_fill := kw_planes_mpl.get('alpha_fill')) is not None:
            kw_planes_mpl.pop('alpha_fill')
        else:
            alpha_fill = 0.3
        zorder = zorder if (zorder := kw_planes_mpl.get('zorder')) is not None else 10
        kw_edge = dict(c='k', lw=1.0, ls='--', alpha=0.5) | kw_planes_mpl
        kw_top = kw_edge | dict(
            lw=2*kw_edge.get('lw'), 
            ls='-', 
            alpha=min(1.0,1.5*kw_edge.get('alpha')),
        )
        if mode == "3d":
            kw_edge = kw_edge | dict(edgecolor=kw_edge.get('c'), alpha=alpha_fill)
            kw_edge.pop('c')

        flag_normal = True
        ic_list = set()
        for ic in results_clustering.query(
                "num_clusters>0"
                "&variance_ratio_normal<=@max_PC3_ratio"  # Quality check by PC3 variance ratio
                "&planarity>=@min_planarity"  # Quality check by planarity
                "&median_angle_diff<=@max_angle_diff"  # Quality check by angular difference
                "&length>=@min_L&width>=@min_W"
                "&n_D>=-@max_nvv"  # Selection by v-comp of normal vector (min_dipping)
            ).index:
            ser_plane = results_clustering.loc[ic]
            df_events = output_events.loc[ic]

            normal_NED = ser_plane[['n_N','n_E','n_D']].values

            box_arr = ser_plane[[f'R{i}_{X}' for i,X in product(range(4),'NED')]].values.reshape(4,3)[:,[1,0,2]]
            box_center = box_arr.mean(axis=0)

            ic_list.add(ic)
            
            if mode == "plotly":
                self.__plot_plane_plotly(
                    ic=ic, 
                    ser_plane=ser_plane,
                    normal_NED=normal_NED,
                    box_arr=box_arr, box_center=box_center,
                    cone_scale=cone_scale,
                    col_plotly=col_plotly,
                    flag_normal=flag_normal,
                    fig=fig,
                )
                flag_normal = False
                
            elif mode == "3d":
                box = Poly3DCollection(
                    [box_arr], 
                    color=cmap(ic/n_clusters), 
                    zorder=zorder-_normalize_min_max(ser_plane.R0_D),
                    **kw_edge
                )
                ax.add_collection3d(box)
                ax.plot(
                    *box_arr[:2].T, 
                    zorder=zorder-_normalize_min_max(ser_plane.R0_D),
                    **kw_top
                )
                
                ax.quiver3D(*box_center, *normal_NED[[1,0,2]], length=cone_scale, zorder=20, lw=1.5, color='k')

            else:  # "2d"
                ax.fill(
                    *box_arr[:,:2].T, 
                    c=cmap(ic/n_clusters), alpha=alpha_fill,
                    zorder=zorder-_normalize_min_max(ser_plane.R0_D)
                )
                ax.plot(
                    *box_arr[[1,2,3,0],:2].T, 
                    zorder=zorder-_normalize_min_max(ser_plane.R0_D),
                    **kw_edge
                )
                ax.plot(
                    *box_arr[[0,1],:2].T, 
                    zorder=zorder-_normalize_min_max(ser_plane.R0_D),
                    **kw_top
                )

    def __plot_plane_plotly(
            self, 
            ic, ser_plane, normal_NED, 
            box_arr, box_center, 
            cone_scale, col_plotly, flag_normal, 
            fig
        ):
        
        fig.add_cone(
            **{k: [v] for k, v in zip('xyz',box_center)},
            **{k: [v] for k, v in zip('uvw',normal_NED[[1,0,2]])},
            hovertemplate=(
                f'i: {ic:03d}<br>'
                '(%{u:4.2f}, %{v:4.2f}, %{w:5.2f})<br>'
                f'strike: {ser_plane.strike:5.1f}°<br>'
                f'dip: {ser_plane.dip:4.1f}°<br>'
                f'L: {ser_plane.length:4.1f} [km]<br>'
                f'W: {ser_plane.width:4.1f} [km]<br>'
                'x: %{x:7.3f}<br>'
                'y: %{y:7.3f}<br>'
                'z: %{z:7.3f}'
            ),
            anchor='tail', opacity=0.4,
            sizemode='scaled', sizeref=cone_scale, showscale=False,
            # colorscale='viridis',
            legendgroup='normal', name='normal vector',
            showlegend=flag_normal, visible='legendonly'
        )

        fig.add_mesh3d(
            **{k: v for k, v in zip('xyz',box_arr.T)}, 
            hovertemplate=(
                f'i: {ic:03d}<br>'
                '%{text}<br>'
                'x: %{x:7.3f}<br>'
                'y: %{y:7.3f}<br>'
                'z: %{z:7.3f}'
            ),
            text=[f'C{ii}' for ii in range(1,5)],
            opacity=0.5, 
            legendgroup=f'cl #{ic:03d}',
            name=f'cl #{ic:03d}',
            showlegend=True,
            # color=color_list[ig]
            color=col_plotly(ser_plane.strike),
        )


@njit("f8[:, :](f8[:, :])", parallel=True)
def compute_distance_absdcos_positions(feature_vector):
    """Compute an distance matrix for clustering. 
    D(f1,f2) = sqrt( |x1-x2|^2 + 2(1-|inner(n1,n2)|) )
    Compute absolute cosine dissimilarity of given normalized vectors.
    
    Parameters:
    -----------
    feature_vector: 2-D ndarray
        A single normalized vector or 2-D array of standardized positions and normalized vectors.
    """
    
    position_arr = feature_vector[:,:3]
    normal_arr = feature_vector[:,3:]
    nsize = len(normal_arr)
    result = np.zeros((nsize,nsize), dtype=np.float64)
    
    for i in prange(nsize):
        pos_i = position_arr[i]
        vec_i = normal_arr[i]
        for j in range(i+1, nsize):
            pos_j = position_arr[j]
            vec_j = normal_arr[j]
            c = 0.0
            for k in range(3):
                c += vec_i[k]*vec_j[k]
            c = max(1-abs(c), 0.0)
            p = 0.0
            for k in range(3):
                p += (pos_i[k]-pos_j[k])**2

            d = np.sqrt(p + 2*c)
            result[i,j] = result[j,i] = d
    
    return result

@dataclass
class FaultHDBSCAN(BaseClustering):
    """Class for clustering events by normal vectors.
    The normal vector is computed from N nearest neighbors.
    
    Note that we use the coordinate system of
    [North, East, Down] positive. Thus, the input data of
    [X,Y,Z] is transformed into [Y,X,Z] (+Z & +D for downward). 
        
    Parameters
    ----------
    max_neighbors_normal_vector: int
        Maximum number of neighbors for calculating normal
    min_neighbors_normal_vector: int
        Minimum number of neighbors for calculating normal
    max_dist_neighbors: float = None
        Maximum distance of neighbors for calculating normal
    min_planarity: float = 1.25
        Minimum planarity (λ2/λ3) of a point to be used
    mode: int, default to 0
        This functions nothing
    
    """

    mode: int = 0
    # max_PC3_ratio_nv: float = 1/3

    rest: dict = field(default_factory=dict)

    def __post_init__(self):
        self.dict_params = dict()
        self.dict_hdbscan_obj = dict()

        self.mode = 0
        if self.mode == 0:
            self._absdcos = compute_distance_absdcos_positions
        else:
            raise ValueError('Invalid `mode` for computing feature distance.')
        
    def _init_HDBSCAN(
            self,
            min_cluster_size: int,
            min_samples: int,
            cluster_selection_epsilon: float,
        ):

        clusterer = HDBSCAN(
            min_cluster_size=min_cluster_size,
            min_samples=min_samples,
            cluster_selection_epsilon=cluster_selection_epsilon,
            metric='precomputed',
            cluster_selection_method='eom',
        )

        return clusterer
    
    def show_computed_params(self):
        """Print the sets of parameters"""
        print(pd.DataFrame.from_dict(self.dict_params, orient='index'))
    
    def get_cluster(self, key):
        params = self.dict_params(key)
        clusterer = self.dict_hdbscan_obj.get(key)

        return params, clusterer

    def fit_predict(
            self,
            min_cluster_size: int = 5,
            min_samples: int = None,
            cluster_selection_epsilon: float = 0.0,
        ):
        """Perform HDBSCAN clustering.
        Parameter set and object are stored in 
        `.dict_params` and `.dict_hdbscan_obj`, respectively.

        Parameters
        ----------
        min_cluster_size: int
            Minimum number of samples to form a cluster
            Must be 3 or more so that λ3 does not fall into zero.
        min_samples: int
        cluster_selection_epsilon: float

        Returns
        -------
        key: int
            A key for `.dict_params` and `.dict_hdbscan_obj`

        Notes
        -----
        - min_cluster_size must be 3 or greater
        - Aware that when min_cluster_size=3, λ3 goes to zero (pure plane)
            In such a case, specify a larger value into `min_points` in `.output_planes()`
        """
        
        if min_cluster_size < 3:
            raise ValueError('min_cluster_size must be less than 3.')
        
        param_dict = dict(
            min_cluster_size=min_cluster_size,
            min_samples=min_samples,
            cluster_selection_epsilon=cluster_selection_epsilon,
        )

        # Avoid re-clustering
        k_trg = len(self.dict_params)
        for k, v in self.dict_params.items():
            if v != param_dict:
                continue
            warnings.warn(
                f"Skip clustering for this parameter set ({k = }):"+"\n"+f"{param_dict}"
            )
            return k
        
        # compute distance matrix
        if "input_data" not in dir(self):
            self.input_data = self._absdcos(self.feature_vector.copy()[self.boolarr_good_nv])
        
        # initialize HDBSCAN object
        clusterer = self._init_HDBSCAN(**param_dict)

        # self.classified_cluster_labels = 
        clusterer.fit_predict(self.input_data)
        self.dict_params.update({k_trg: param_dict})
        self.dict_hdbscan_obj.update({k_trg: clusterer})

        return k_trg

    def output_planes(self, key, min_points=5):
        """Output clustering results of extracted planes and events

        Parameters
        ----------
        key: int
            The key index of clustering
        min_points: int
            Mimimum number of points in a cluster

        Returns
        -------
        results_clustering: pandas.DataFrame
            Dataframe for extracted fault parameters
        output_events: pandas.DataFrame
            Dataframe for hypocenter cluster indices
        
        """

        labels = self.dict_hdbscan_obj[key].labels_

        return super().output_planes(labels=labels, min_points=min_points)

    def plot_events(
        self,
        data=None,
        labels=None,
        includes_normal=True,
        mode='2d', 
        color_normal=True,
        s=2, 
        lw=1.5,
        size_vector=1.,
        cmap=None,
        cbar_rect=[0.8,0.2,0.15,0.15],
        alpha_event=0.7,
        quiverkey_kw=None,
        view_kw=None,
        fig_kw=None, 
        subplot_kw=None,
        fig=None,
        ax=None,
    ):
        """Plot events and normal vectors

        Parameters
        ----------
        data: ndarray, default to None
            Data points to be ploted
            If None, original input data will be plotted.
        labels: ndarray, default to None
            Labels of data points
        includes_normal: bool, default to True
            Whether to include normal vectors.
        mode: str, default to `'2d'`
            Type of projection. Select one from
            `'2d'`, `'3d'`,  and `'cartopy'`
            `'plotly'` is not currently supported.
        color_normal: bool, default to True
            Whether colors are based on normal vectors (True)
            or based on apparent strike (False)
        s: float
            Size of scatter plot
        lw: float
            Linewidth
        size_vector: float
            Size of normal vectors
        cmap: str or mcolors.Colormap, default to None
            cmap of colorbar. 
            If None, `DEFAULT_CYCLIC_CMAP` is used.
        cbar_rect: list or ndarray
            Position and size of colorbar
        alpha_event: float
            Opacity of scatter plot
        quiverkey_kw: dict
            kwargs for `ax.quiverkey()`
        view_kw: dict
            kwargs for `ax.view_init()`
        fig_kw: dict
             kwargs for `plt.figure()`
        subplot_kw: dict
            kwargs for `fig.add_subplot()`
            Specify `projection` when mode=`cartopy`.
        fig: plt.Figure, default to None
            Specify an existing Figure object or newly create
        ax: plt.Axes, default to None
            Specify an existing AXes object or 
            newly create from the figure object.

        Returns
        -------
        fig: plt.Figure
        """
        if mode not in ["2d","3d","cartopy","plotly"]:
            raise ValueError(f"Invalid input '{mode}' for `mode`.")
        
        norm = plt.Normalize(0, 360)

        if mode == 'plotly':
            raise ValueError("`plotly` is not currently supported.")
        else:
            if mode == 'cartopy':
                import cartopy.crs as ccrs
                pc = ccrs.PlateCarree()
                kw_pc = dict(transform=pc)
            else:
                kw_pc = dict()

            if isinstance(ax, plt.Axes):
                fig = ax.figure
            if fig is None:
                default_fig_kw = dict(figsize=(9,8), dpi=100)
                if fig_kw:
                    default_fig_kw.update(fig_kw)
                fig = plt.figure(**default_fig_kw)

            if not isinstance(ax, plt.Axes):
                default_subplot_kw = dict()
                if subplot_kw:
                    default_subplot_kw.update(subplot_kw)
                
                if mode == "cartopy":
                    if (subplot_kw is None) or (subplot_kw.get('projection') is None):
                        projection = pc
                    else:
                        projection = subplot_kw.get('projection')
                elif mode == "3d":
                    projection = "3d"
                else:
                    projection = None
                
                ax = fig.add_subplot(
                    projection=projection, 
                    **default_subplot_kw
                )

            if not isinstance(cmap, mcolors.Colormap):
                if isinstance(cmap, str):
                    cmap = plt.get_cmap(cmap)
                else:
                    cmap = plt.get_cmap(DEFAULT_CYCLIC_CMAP)

            XYZ_data = self.XYZ_normal_vectors_kneighbors[self.boolarr_good_nv]
            if labels is not None:
                XYZ_data = XYZ_data[labels>=0]
        
            if mode == "3d":
                ndims = 3
                data = XYZ_data#[self.classified_cluster_labels>=0]

                ax.set_zlabel('Depth [km]')
                ax.invert_zaxis()
            else:
                ndims = 2
                if data is None:
                    data = XYZ_data[:,[0,1,3,4]] #[self.classified_cluster_labels!=-1][:,[0,1,3,4]]



            azimuth_data = (np.rad2deg(np.angle(list(map(lambda X: X[1]+1j*X[0] , data[:,ndims:]))))-[90,0][color_normal])%360

            ax.scatter(
                *data[:,:ndims].T, 
                s=s, #c=[color],
                c=azimuth_data,
                edgecolors='none',
                cmap=cmap, norm=norm,
                alpha=alpha_event,
                **kw_pc
            )

            if includes_normal:
                if mode == "3d":
                    ax.quiver(
                        *data.T, 
                        color=plt.get_cmap(cmap)(norm(azimuth_data)),
                        length=size_vector, 
                        zorder=20, 
                        lw=lw, 
                        alpha=0.5
                    )
                elif mode in ["2d", "cartopy"]:
                    Q = ax.quiver(
                        *data.T, 
                        color=plt.get_cmap(cmap)(norm(azimuth_data)),
                        scale=1./size_vector,
                        angles='xy',scale_units='xy',
                        zorder=20, 
                        lw=lw, 
                        alpha=0.5,
                        **kw_pc
                    )
                    default_quiverkey_kw = dict(
                        X=0.8, Y=0.95, U=1.2, 
                        label='Normal vector\n$|n|=1$', 
                        labelpos='E', coordinates='axes',
                    )
                    if quiverkey_kw:
                        default_quiverkey_kw.update(quiverkey_kw)
                    ax.quiverkey(Q, **default_quiverkey_kw)

            if mode != "cartopy":
                ax.set(
                    aspect='equal',
                    xlabel='Easting [km]', ylabel='Northing [km]',
                )
            
            if cbar_rect:
                r, theta = np.meshgrid(np.linspace(0.8, 1, 2), np.linspace(0, 2*np.pi, 1000))

                ax_cb = ax.inset_axes(cbar_rect, projection='polar')
                ax_cb.grid(False)
                ax_cb.spines['polar'].set_visible(False)
                ax_cb.contourf(theta, r, theta, levels=1000, cmap=cmap)
                ax_cb.text(
                    0,0,
                    ['Strike','Normal\nvector'][color_normal]+r' [$\degree$]',
                    ha='center',va='center'
                )
                ax_cb.set(
                    rlim=(0,1), rticks=[],
                    aspect='equal',
                    theta_direction='clockwise',
                    theta_zero_location='N',
                    xticks=np.deg2rad(np.arange(0,360,30)),
                )
                fig.axes.append(ax_cb)

            if (mode == "3d") and view_kw:
                ax.view_init(**view_kw)
            
        return fig

    def plot_faults(
            self,
            key=None,
            data=None,
            mode="2d",
            max_angle_diff=30., 
            max_PC3_ratio=1/3,
            min_planarity=1.0,
            min_points=5,
            min_dipping=0.,
            min_L=0.0,
            min_W=0.0,
            size=1., 
            kw_planes_mpl=dict(c='k', lw=1., ls='--', alpha=0.5),
            cmap='jet', 
            cone_scale=1.0, 
            plotly_layout=dict(),
            fig_kw=None,
            subplot_kw=None,
            fig=None, 
            ax=None,
            **view_kwargs
        ):
        """Plot extracted fault planes

        Parameters
        ----------
        key: int
            Key index for clustering
        data: ndarray, default to None
            Data points to be ploted
            If None, original input data will be plotted.
        mode: str, default to `'2d'`
            Type of projection. Select one from
            `'2d'`, `'3d'`, and `'plotly'`
        max_angle_diff: float, default to 30.0
            Maximum median angle difference between
            fault normal and point-cloud normal vectors.
            Specify 90.0 to plot all the clusters 
        max_PC3_ratio: float, default to 1/3
            Maximum PC3 ratio to plot
        min_planarity: float, default to 1.0
            Miminum planarity to plot
        min_points: int
            Mimimum number of points in a cluster
        min_dipping: float
            Mimimum dipping angle
        min_L: float
            Mimimum length of a fault plane
        min_W: float
            Mimimum width of a fault plane
        size: float
            Size of scatter plot
        kw_planes_mpl: dict
            kw dict in mpl plot for planes
        cmap: str or mcolors.Colormap, default to None
            cmap of colorbar. 
            If None, `DEFAULT_CYCLIC_CMAP` is used.
        cone_scale: float
            Size of normal vectors
        plotly_layout: dict
            `layout` argument for `plotly.graph_objects.Figure()`
        fig_kw: dict
             kwargs for `plt.figure()`
        subplot_kw: dict
            kwargs for `fig.add_subplot()`
        fig: plt.Figure, default to None
            Specify an existing Figure object or newly create
        ax: plt.Axes, default to None
            Specify an existing AXes object or 
            newly create from the figure object.

        Returns
        -------
        fig: plt.Figure
        """
        results_clustering, output_events = self.output_planes(key, min_points=min_points)

        fig = super().plot_faults(
            data=data,
            mode=mode,
            max_angle_diff=max_angle_diff, 
            max_PC3_ratio=max_PC3_ratio, 
            min_planarity=min_planarity,
            min_dipping=min_dipping,
            min_L=min_L,
            min_W=min_W,
            size=size, 
            kw_planes_mpl=kw_planes_mpl,
            cmap=cmap, 
            cone_scale=cone_scale, 
            plotly_layout=plotly_layout,
            fig_kw=fig_kw,
            subplot_kw=subplot_kw,
            fig=fig, 
            ax=ax,
            results_clustering=results_clustering,
            output_events=output_events,
            **view_kwargs
        )

        return fig
    
    def save_model(self, filename, key, min_points=5):
        """Save fault model in CSV format
        
        Parameters
        ----------
        filename: str
            Filename (path to the file)
        key: int
            Key index for clustering
        min_points: int
            Mimimum number of points in a cluster

        Returns
        -------
        filename+'_planes.csv' <- results_clustering
        filename+'_events.csv' <- output_events
        """
        filename = Path(filename)
        results_clustering, output_events = self.output_planes(key=key, min_points=min_points)

        results_clustering.to_csv(filename.parent/(filename.stem+'_planes.csv'), index=True)
        output_events.to_csv(filename.parent/(filename.stem+'_events.csv'), index=True)

    def plot_dendrogram(self, key, ax=None, recursionlimit=None, **kwargs):

        if recursionlimit is not None:
            sys.setrecursionlimit(recursionlimit)

        ax = self.dict_hdbscan_obj.get(key).single_linkage_tree_.plot(axis=ax, **kwargs)

        ax.set(
            ylabel=r'distance $\epsilon$'
        )

        return ax
