# fnvc/synthetic.py
# See LICENSE file for full license details.
# Copyright (c) 2024, National Institute of Advanced Industrial Science and Technology (AIST). All rights reserved.


import numpy as np


def synthesize_cluster(
    n_points, 
    normal_vector_NED, 
    planar_shape='uniform', 
    scale=0.1, 
    strike_lim=[-2,2], 
    dip_lim=[-1,1],
    kw_circ=dict(
        rmax=1.0
    ),
    returns_as_END=True,
    rseed=None,
):
    """Synthesize point cloud data around a given planar shape.
    The origin of data points is [0,0,0].

    Parameters:
    -----------
    n_points: int
        The number of data points to be synthesized.
    normal_vector_NED: ndarray
        Ndarray of the normal vector in the [N,E,D] system.
    planar_shape: str, default to 'uniform'
        Shape of the plane on which data points are distributed.
        ['uniform', 'circular']
    scale: float, default to 0.1
        Standard deviation of Gaussian noise added to the data points
        If 0.0, noise-free data is returned.
    """

    if rseed is None or not isinstance(rseed, int):
        rseed = np.random.randint(2**16)

    rng = np.random.default_rng(rseed)

    # Strike: π/2 counter-clockwise of the normal vector
    strike_vector = np.array([normal_vector_NED[1],-normal_vector_NED[0],0])
    strike_vector /= np.linalg.norm(strike_vector)

    # Dip: [-nd*nn, -nd*ne, 1-nd*nd]
    dip_vector = -normal_vector_NED[2]*normal_vector_NED
    dip_vector[2] += 1
    dip_vector /= np.linalg.norm(dip_vector)
    
    if planar_shape == 'circular':
        r = rng.uniform(0, kw_circ.get('rmax'), size=n_points)
        phi = rng.uniform(0, 2*np.pi, size=n_points)
        xi = r*np.cos(phi)
        eta = r*np.sin(phi)
    elif planar_shape == 'gaussian':
        raise Exception('`planar_shape`=="gaussian" has not been supported yet.')
    else:
        xi = rng.uniform(*strike_lim,size=n_points)
        eta = rng.uniform(*dip_lim,size=n_points)
        
    datapoints = np.outer(xi,strike_vector) + np.outer(eta,dip_vector)
    if scale > 1e-8:
        datapoints += rng.normal(scale=scale, size=(n_points,3))
    
    # NED to END
    if returns_as_END:
        datapoints = datapoints[:,[1,0,2]]
    
    return datapoints


def normalize(v):
    v = v/np.linalg.norm(v)
    v *= 1-2*int(v[2]>0)
    return v
