# fnvc/utils.py
# See LICENSE file for full license details.
# Copyright (c) 2024, National Institute of Advanced Industrial Science and Technology (AIST). All rights reserved.


import string

import numpy as np
import matplotlib.colors as mcolors
from sklearn.decomposition import PCA
from pyproj import Transformer


def init_transformer(epsg_local):
    """Initialize `pyproj.Transformer` from a local epsg code.
    """
    transformer_4326_xxxx = Transformer.from_crs(
        "epsg:4326", epsg_local, always_xy=True
    )
    transformer_xxxx_4326 = Transformer.from_crs(
        epsg_local, "epsg:4326", always_xy=True
    )

    return transformer_4326_xxxx, transformer_xxxx_4326


def normalize_min_max(values, min, max):
    if max <= min:
        return values * 0
    return (values - min) / (max - min)


def compute_each_normal_vector(locations, returns_EVR=False):
    """Compute the normal vector (PC3 eigen vector).
    Normal vector is defined as upward direction, but we use the coordinate system of
    [North, East, Down] positive. So, Z component of the normal vector must be negative.

    Parameters:
    -----------
    locations: ndarray
        Array of event locations (NED)
    """

    ## Using SVD in `numpy.linalg`
    # 2x faster than scikit-learn PCA
    # _, _, components = np.linalg.svd(locations-locations.mean(axis=0))
    # n_dot_calc = components[2]

    pca = PCA(n_components=3)
    pca.fit(locations)

    # PC3 eigen vector: Normal vector of the plane (PCA1, PCA2)
    n_dot_calc = pca.components_[2]
    # which is same as the cross product of PCA1 & PCA2
    # n_dot_calc = np.cross(*pca.components_[:2])

    # Eigen value variance ratio
    variance_ratio = pca.explained_variance_ratio_
    
    # Reverse the polarity so that Z comp is to be negative
    n_dot_calc *= 1 - 2*int(n_dot_calc[2]>=0)

    if returns_EVR:
        return np.hstack([n_dot_calc, variance_ratio])
    else:
        return n_dot_calc


def normal2others(normal_vector):
    
    strike_vector = np.array([normal_vector[1],-normal_vector[0],0.])
    strike_vector /= np.linalg.norm(strike_vector)
    dip_vector = np.cross(-normal_vector, strike_vector)

    return strike_vector, dip_vector


def convert_strikedip2normal(strike, dip):
    strike = np.deg2rad(strike)
    dip = np.deg2rad(dip)

    normal_vector =  np.array([
        -np.sin(dip)*np.sin(strike), 
        np.sin(dip)*np.cos(strike), 
        -np.cos(dip)
    ]).T

    dip_vector =  np.array([
        -np.cos(dip)*np.sin(strike), 
        np.cos(dip)*np.cos(strike), 
        np.sin(dip)
    ]).T

    return normal_vector, dip_vector


def transform_into_plane(data_in_NED, normal_vector, strike_vector, dip_vector):
  
    P_inv_T = np.linalg.inv(np.vstack([strike_vector,dip_vector,-normal_vector]))
    
    outdata_in_NED = data_in_NED @ P_inv_T
    return outdata_in_NED

def compute_fault_param(points):

    normal_vector = compute_each_normal_vector(points)
    strike_vector, dip_vector = normal2others(normal_vector)
    
    points_in_SDN = transform_into_plane(
        points, 
        normal_vector, 
        strike_vector, 
        dip_vector
    )
    
    normal_mean = np.mean(points_in_SDN[:,2])  # mean along normal axis
    variance = np.var(points_in_SDN, axis=0)   # variance along normal axis
    planarity = variance[1]/variance[2]
    pc3_variance_ratio = variance[2]/np.sum(variance)
    
    xi_min, eta_min = points_in_SDN[:,:2].min(axis=0)
    xi_max, eta_max = points_in_SDN[:,:2].max(axis=0)
    fault_length = np.round(xi_max-xi_min, 3)
    fault_width = np.round(eta_max-eta_min, 3)
    
    box_arr = np.array([
        [xi_min, eta_min],
        [xi_max, eta_min],
        [xi_max, eta_max],
        [xi_min, eta_max],
    ]) @ np.array([strike_vector, dip_vector]) + normal_mean*(-normal_vector)
    
    box_arr = np.round(box_arr, 3)
    
    return (
        box_arr, 
        fault_length, 
        fault_width, 
        normal_vector,
        strike_vector, 
        dip_vector, 
        variance, 
        planarity,
        pc3_variance_ratio
    )

def fix_aspectratio(fig, only_visible=True):
    """This function fixes the aspect ratio for plotly plot.
    Note that plotly must be importable on your environment.
    """
    import plotly.graph_objects as go
    if not isinstance(fig, go.Figure):
        raise TypeError('`fig` must be `plotly.graph_objects.Figure`')
        
    x_range, y_range, z_range = [], [], []
    for trace in fig.data:
        if only_visible:
            if trace.visible in ["legendonly", False]:
                continue
        if 'x' in trace and 'y' in trace and 'z' in trace:
            if len(x_range) == 0:
                x_range = [np.nanmin(trace.x), np.nanmax(trace.x)]
                y_range = [np.nanmin(trace.y), np.nanmax(trace.y)]
                z_range = [np.nanmin(trace.z), np.nanmax(trace.z)]
                continue
            x_range = [min(np.nanmin(trace.x),x_range[0]), max(np.nanmax(trace.x),x_range[1])]
            y_range = [min(np.nanmin(trace.y),y_range[0]), max(np.nanmax(trace.y),y_range[1])]
            z_range = [min(np.nanmin(trace.z),z_range[0]), max(np.nanmax(trace.z),z_range[1])]
            
    data = np.vstack([x_range, y_range, z_range]).T
    aspectratio = data.max(axis=0)-data.min(axis=0)
    aspectratio = {s: float(d)/np.max(aspectratio) for s, d in zip('xyz',aspectratio)}

    fig.update_layout(
        scene=dict(
            aspectmode='manual',
            aspectratio=aspectratio,
        )
    )

    return fig

def _initialize_CSS4_COLORS():
    """Extract CSS4_COLORS w.o. white-based colors.
    """
    # colorlist_plot = list(mcolors.CSS4_COLORS.values())
    # [
    #     colorlist_plot.remove(c) 
    #     for c in colorlist_plot 
    #     if all([c[2*j+1:2*j+3].startswith(('F','E')) for j in range(3)])
    # ]
    colorlist_plot = [
        c for c in list(mcolors.CSS4_COLORS.values())
        if not all([c[2*j+1:2*j+3].startswith(tuple(string.ascii_uppercase[:6])) for j in range(3)])
    ]
    np.random.shuffle(colorlist_plot)
    
    return colorlist_plot